﻿
namespace HeuristicLab.Problems.Instances.TSPLIB {
  /// <summary>
  /// This class is used to randomly generate PTSP instances given the TSPLIB instances.
  /// An own implementation of a RNG was used in order to avoid possible implementation changes
  /// in future .NET versions which would result in entirely different instances.
  /// </summary>
  internal class MarsagliaRandom {
    /*
     * S = 2111111111*X[n-4] + 1492*X[n-3] + 1776*X[n-2] + 5115*X[n-1] + C
     * X[n] = S modulo 2^32
     * C = floor(S / 2^32)
     * 
     */
    private readonly uint[] mem = new uint[4];
    private uint c;

    public MarsagliaRandom(uint s) {
      Seed(s);
    }

    private uint Next() {
      unchecked {
        ulong wsum = 2111111111 * mem[0]
                    + 1492 * mem[1]
                    + 1776 * mem[2]
                    + 5115 * mem[3]
                    + c;

        mem[0] = mem[1];
        mem[1] = mem[2];
        mem[2] = mem[3];
        mem[3] = (uint)wsum;
        c = (uint)(wsum >> 32);
        return mem[3];
      }
    }

    public double NextDouble() {
      return (double)Next() / uint.MaxValue;
    }

    public void Seed(uint seed) {
      int i;
      for (i = 0; i < mem.Length; i++) {
        unchecked {
          seed = seed * 31294061 + 1;
        }
        mem[i] = seed;
      }
      unchecked {
        c = seed * 31294061 + 1;
      }
    }
  }
}
